package evote.admin;

import java.io.*;
import java.math.BigInteger;
import java.net.*;
import java.security.*;

import evote.crypto.*;
import evote.commissioner.complaints.*;
import evote.lang.*;
import evote.network.message.*;
import evote.network.server.*;
import evote.registrar.*;
import evote.util.constants.*;
import evote.util.genrand.*;
import evote.identity.*;
import evote.election.*;
import evote.question.*;
import evote.commissioner.*;

/**
  The AdminServer is responsible for signing voters voters, after confirming
  the voters right to vote.  It implaments the ConnectionHandler interface.
  The server keeps track of the number of attemtped connections as well as
  the number of confirmed signatures.  It has it's own public key set, and
  access to the regitration files of the voters.
*/
public class AdminServer implements ConnectionHandler {
  
    private Election myElection;
    private UserList myUserList;
    private UserListWatcher watcher;

    /* File with people who already voted */
    private File alreadyVotedFile;
    private PrintWriter alreadyVotedStream;
  
    String my_name = "Admin";
    Server my_server;
    int total_connection_count = 0;
    int total_confirmed_count = 0;
    VoteMessage msg;
    public GenRand RNG;

    /* add in Private Key field here */
    RsaEncryption adminSigner;


    /**
       This is the only constructor for the AdminServer.
    */
    public AdminServer(Election elec) {
	try {
	    myElection= elec;
	    Constants.init(myElection);

	    /* Set up all the keys
	     */
	    adminSigner= new RsaEncryption(Constants.adminKeys);

	    System.out.println("got all parameters");
	    myUserList= myElection.getAllUsers();
	    System.out.println("loaded up users");
	    RNG = new GenRand("evote.util.genrand.MouseGenerator");
	    RNG.seed();

	    /* Check who has already voted */
	    alreadyVotedFile= new File("ALREADY_VOTED.txt");
      
	    String votedList= "";
	    if (alreadyVotedFile.exists()) {
		try {
		    FileReader fread= new FileReader(alreadyVotedFile);
		    BufferedReader buffread= new BufferedReader(fread);
	  
		    System.out.println("Parsing List of Voters who have already voted...");

		    String line;
		    while((line= buffread.readLine()) != null) {
			User tempUser= myUserList.findUser(line);
			System.out.print("User -"+line+"-");
			if (tempUser != null) {
			    votedList+= tempUser.getUserName() + "\n";
			    tempUser.voted();
			    System.out.println("FOUND");
			} else {
			    System.out.println("NOT FOUND");
			}
		    }
		    buffread.close();
		} catch(Exception exc) {
		}
		System.out.println("Done Parsing");
	    }
      
	    // for now, check who voted.
	    // watcher= new UserListWatcher(myUserList);
      
	    FileOutputStream fos= new FileOutputStream(alreadyVotedFile);
	    alreadyVotedStream= new PrintWriter(fos);
	    alreadyVotedStream.println("ALREADY VOTED LIST");
	    alreadyVotedStream.println("==================\n");
	    alreadyVotedStream.flush();

	    /* reappend all the ones who have already voted */
	    alreadyVotedStream.print(votedList);
	    alreadyVotedStream.flush();

	    /* Start Server and all */
	    my_server= new SecureServer(my_name, Constants.adminReceivePort, this, Constants.adminNumThreads, Constants.adminQueueSize, Constants.adminKeys, RNG, false);
	    ((SecureServer)my_server).setAcceptInsecure(false);

	    /* Set times */
	    my_server.setTimes(Constants.adminStartDate,Constants.adminEndDate);
	    my_server.start();
	}
	catch(Exception exp) {
	    System.err.println("Couldn't start up admin server: " + exp.toString());
	    exp.printStackTrace();
	}
    }
  
    public void startDemo() {
	my_server.startDemo();
    }

    public void stopDemo() {
	my_server.stopDemo();
    }

    /**
       Increments the count of the number of connections.
    */
    private synchronized void increase_connection_counter() {
	total_connection_count += 1;
    }

    /**
       Increments the count of the number of signatures given.
    */
    private synchronized void increase_confirmed_counter() {
	total_confirmed_count += 1;
    }

    /**
       This is the startup sequence.
    */
    public static void main(String argv[]) {
    
	Election myElec= null;

	try {
	    ElectionLoader eLoader= new ElectionLoader(argv);
	    myElec= eLoader.getElection();
      
	    new AdminServer(myElec);
	    System.out.println("Admin Server starting ....");
	} catch (Exception exc) {
	    System.out.println("Error happened: " + exc.toString());
	}
    }

    /**
       This is the method used to perform the Admins task in the election protocol.
       Upon receving a message it first check that it is of type VoteMessage.  Next,
       the voter's name and password are confirmed.  If these are correct the
       blinded committed vote gets signed and sent back.
    */
    public void handleConnection(Connection c) {

	try {
	    my_server.setStatus("RECEIVED\nCONNECTION");

	    // System.out.println("GOT CONNECTION");
	    increase_connection_counter();

	    Message incoming= c.receiveMessage();
	    // System.out.println("RECEIVED MESSAGE");

	    if (incoming instanceof LogonMessage) {
		my_server.setStatus("USER\nLOGGING ON");

		// System.out.println("IT'S A LOGON MESSAGE");
		VoterIdentityObject vido= ((LogonMessage)incoming).getIdentity();
		System.out.println("looking for user " + vido.getName());
		User thisUser= myUserList.findUser(vido.getName());
		System.out.println("got user");
		if (thisUser == null) {
		    my_server.setStatus("Incorrect\nUser");
		    try {
			c.sendMessage(new ErrorMessage("You are not authorized to vote!"));
		    } catch (Exception e) {
		    }
		    throw(new Exception("Voter does not Exist"));
		}
	
		if (!thisUser.verify(vido.getPassword())) {
		    my_server.setStatus("Password\nIncorrect!");

		    try {
			c.sendMessage(new ErrorMessage("Your Password is incorrect!"));
		    } catch (Exception e) {
		    }
		    throw(new Exception("Voter Password Incorrect"));
		}

		// get Question List for user
		QuestionList userQL= myElection.makeUserQuestionList(thisUser);
		try {
		    System.out.println("Number of questions: " + userQL.getNumQuestions());
		} catch(Exception exc) {
		    System.err.println(exc.toString());
		}
		// int[] questionLists= myElection.getQuestionListNumbers(thisUser);

		// make question list message, sign question list
		QuestionListMessage qlm= new QuestionListMessage(userQL);

		// get question list bytes, sign it
		ByteArrayOutputStream baos= new ByteArrayOutputStream();
		ObjectOutputStream oos = new ObjectOutputStream(baos);
		oos.writeObject(userQL);
		byte[] QLBytes= baos.toByteArray();
	
		// hash it using commitment scheme
		ShaCommitmentKeys sha_keys= new ShaCommitmentKeys(new BigInteger("0"), new BigInteger("0"));
		ShaCommitment sha= new ShaCommitment(sha_keys);
		byte[] hashedQLBytes= sha.commit(QLBytes);
		byte[] signedQLBytes= adminSigner.sign(hashedQLBytes);

		my_server.setStatus("Sending\nQuestionlist");

		qlm.setSignature(signedQLBytes);
		c.sendMessage(qlm);

		// by returning, we close the connection automatically
		return;
	    }

	    my_server.setStatus("Incoming\nVote");
	    System.out.println("not a logon message");

	    msg = (VoteMessage)incoming;
	    String voteName = (msg.getVoterID()).getName();
	    String votePwd = (msg.getVoterID()).getPassword();

	    User thisUser= myUserList.findUser(voteName);
	    if (thisUser == null) {
		try {
		    c.sendMessage(new ErrorMessage("You are not an authorized voter"));
		} catch (Exception e) {
		}
		throw(new Exception("Voter does not Exist"));
	    }

	    System.out.println("user exists");

	    if (!thisUser.verify(votePwd)) {
		try {
		    c.sendMessage(new ErrorMessage("Your Password is incorrect!"));
		} catch (Exception e) {
		}
	
		throw(new Exception("Incorrect Password"));
	    }

	    my_server.setStatus("User\nVerified");
	    System.out.println("correct password");

	    /* We lock on the User */
	    synchronized(thisUser) {
		if (thisUser.hasVoted()) {
		    my_server.setStatus("Already\nVoted!");
		    try {
			c.sendMessage(new ErrorMessage("You have already voted!"));
		    } catch (Exception e) {
		    }

		    System.out.println("already voted");
		    throw(new Exception("Already Voted"));
		}

		my_server.setStatus("Signing\nVote");

		try {
		    (msg.getVote()).adminSign(adminSigner);
		} catch (Exception exp) {
		    try {
			c.sendMessage(new ErrorMessage("Serious Configuration Problem. Please contact the Administrator ASAP and notify him of this problem"));
		    } catch (Exception e) {
		    }
		    System.out.println("coudln't sign");
		    throw(new Exception("Exception signing " + voteName + "'s vote: " + exp.toString()));
		}

		my_server.setStatus("Marking\nUser Voted");

		try {
		    thisUser.voted();
		    /* Log that this user voted */
		    alreadyVotedStream.println(thisUser.getUserName());
		    alreadyVotedStream.flush();
	  
		    System.out.println("sending message back");

		    my_server.setStatus("Sending\nSignature");
		   
		    my_server.incrementCount();

		    c.sendMessage(msg);
		    increase_confirmed_counter();
		    // my_server.log(new LogEntry(LogEntry.NORMAL_LOG, voteName + "\n" + (new BigInteger((msg.getVote()).getBlindedBytes())).toString() + "\n" + (new BigInteger((msg.getVote()).getAdminSignature())).toString()));
	  
		    // watcher.update(thisUser);
	  
		} catch(Exception exp) {
		    System.out.println("problem: " + exp.toString());
		    exp.printStackTrace();
		    throw(new Exception("Can't send " + voteName + "'s signed vote back to back. --> " + exp.toString()));
		}
	    }
 
	}
	catch (Exception exp) {
	    System.err.println("problem");
	    exp.printStackTrace();
	    Complaint comp = new Complaint(my_name, (c.getInetAddress()).toString(), 0, exp.toString(), RNG);
	    ComplaintMessage compMsg = new ComplaintMessage(comp, adminSigner);
	    my_server.log(new LogEntry(LogEntry.WARNING_LOG, my_name + "\n" + (c.getInetAddress()).toString() + "\n" + exp.toString()));
	    try {
		SecureConnection commisConnection = new SecureConnection(Constants.commisServer, Constants.commisReceivePort, RNG, Constants.commisKeys);
		commisConnection.sendMessage(compMsg);
		c.sendMessage(compMsg);
	    } catch (Exception e) {
		my_server.log(new LogEntry(LogEntry.SERIOUS_LOG, "Can't connect to commissioner."));
	    }
	} finally {
	    my_server.clearStatus();
	}
    }

    /**
       Method used to handle conncetion errors.
    */
    public void handleError(ConnectionError ce) {
	try {
	    int error_type = 1;
	    Complaint comp = new Complaint(my_name, "", error_type, ce.toString(), RNG);
	    ComplaintMessage compMsg = new ComplaintMessage(comp, adminSigner);
	    my_server.log(new LogEntry(LogEntry.WARNING_LOG, my_name + "\n" + "" + "\n" + error_type + "\n" + ce.toString()));
	    SecureConnection commisConnection = new SecureConnection(Constants.commisServer, Constants.commisReceivePort, RNG, Constants.commisKeys);
	    commisConnection.sendMessage(compMsg);
	} catch (Exception exp) {
	    my_server.log(new LogEntry(LogEntry.SERIOUS_LOG, "Can't connect to commissioner."));
	}
    }

    /* server is shutting down
     */
    public void serverShutdown() {
	try {
	    alreadyVotedStream.flush();
	    alreadyVotedStream.close();
	} catch(Exception exc) {
	}
    }

}



