/* Electronic Voting Project
 * Ben Adida(ben@mit.edu)
 * This is the class that performs the actual counting of votes
 */

package evote.counter;

import java.io.*;
import java.util.Enumeration;
import java.util.Vector;
import evote.lang.*;
import evote.counter.*;
import evote.election.*;
import evote.question.*;

/**
 * The Vote Counter
 *
 * Given a number of choices, this is the object that performs
 * the counting of the choices.
 *
 * @see VoteCounterDisplay
 */
public class VoteCounter {
  private Election my_election;
  private boolean votes_counted= false;
  
  /* The dimensions for vote_counts is [questionListNum][questionNum][choiceNum] */
  private int[][][] vote_counts;
  /* The dimensions for write_ins is [questionListNum][questionNum] */
  private Vector[][] write_ins;

  private int num_erroneous_votes=0;

  /**
   * Create the Vote Counter
   * 
   * @argument e The election involved
   */
  public VoteCounter(Election e) {
    my_election= e;

    QuestionList[] questionLists= e.getQuestionLists();
    
    vote_counts= new int[questionLists.length][][];
    write_ins= new Vector[questionLists.length][];
    int num_questions, num_choices;
    Question currentQuestion;
    for (int i=0; i<questionLists.length; i++) {
      num_questions= questionLists[i].getNumQuestions();
      vote_counts[i]= new int[num_questions][];
      write_ins[i]= new Vector[num_questions];
      for (int j=0; j<num_questions; j++) {
	currentQuestion= questionLists[i].getQuestion(j);
	num_choices= currentQuestion.getNumAnswers();
	vote_counts[i][j]= new int[num_choices];
	write_ins[i][j]= new Vector();
	for (int k=0; k<num_choices; k++) {
	  vote_counts[i][j][k]= 0;
	}
      }
    }
  }

  /**
   * Count the votes 
   *
   * @argument choice_enum The votes to count
   */
  public synchronized void countVotes(Enumeration vote_enum) {
    Vote currentVote;
    QuestionList ql;
    Question currentQuestion;
    Choice c;

    /* If Votes have already been counted, then don't count anymore*/
    if (votes_counted)
      return;
    
    votes_counted= true;

    while(vote_enum.hasMoreElements()) {
      try {
	currentVote= (Vote)vote_enum.nextElement();
	ql= currentVote.getQuestionList(); 
	c= currentVote.getChoice();
      } catch (Exception exc) {
	num_erroneous_votes += 1;
	System.out.println("VOTE COUNTER: Erroneous Vote");
	continue;
      }

      // System.out.println("Considering one VOTE");

      int i=0;
      try {
	for (i=0; i<c.getSize(); i++) {
	  currentQuestion= ql.getQuestion(i);
	  try {
	    int ch= c.getChoice(i+1);
	    // System.out.println("Choice of question " + i + " is " + ch);
	    if (ch==-1) {
	      /* No choice for this question */
	      continue;
	    }
	    // System.out.println("Got choice " + ch + " for question " + i);
	    vote_counts[currentQuestion.getQuestionListID()][currentQuestion.getQuestionID()][ch] += 1;
	  }
	  catch(WriteInException wie) {
	    try {
	      write_ins[currentQuestion.getQuestionListID()][currentQuestion.getQuestionID()].addElement(c.getWriteInChoice(i+1));
	    }
	    catch(ChoiceException ce) {
	      /* Should not happen */
	    }
	  }
	}
      }
      catch(ArrayIndexOutOfBoundsException aioobe) {
	/* Undo count */
	System.out.println("Array index outofbounds");
	for (int j=0; j<i; j++) {
	  currentQuestion= ql.getQuestion(j);
	  try {
	    int ch= c.getChoice(j+1);
	    if (ch==-1)
	      continue;
	    vote_counts[currentQuestion.getQuestionListID()][currentQuestion.getQuestionID()][ch] -= 1;
	  }
	  catch(WriteInException wie) {
	    try {
	      write_ins[currentQuestion.getQuestionListID()][currentQuestion.getQuestionID()].removeElement(c.getWriteInChoice(j+1));
	    }
	    catch(ChoiceException ce) {
	      /* Should Not Happen */
	    }
	  }
	}
	num_erroneous_votes += 1;
      }

    }

    // output the results
    try {
      FileOutputStream fos= new FileOutputStream("VOTE-COUNT.txt");
      PrintStream dos= new PrintStream(fos);
      QuestionList[] qLists= my_election.getQuestionLists();

      for (int i=0; i<vote_counts.length; i++) {
	dos.println("Questionlist #" + (i+1));

	for (int j=0; j<vote_counts[i].length; j++) {
	  dos.println("Question #" + (j+1));

	  currentQuestion= qLists[i].getQuestion(j);
	  dos.println(currentQuestion.getQuestion() + "\n");

	  String[] answers= currentQuestion.getAnswers();

	  for (int k=0; k<vote_counts[i][j].length; k++) {
	    dos.println(answers[k] + " : " + vote_counts[i][j][k]);
	  }

	  dos.println("\n\n");
	}
      }
    } catch(Exception exc) {
    }
			
  }

  /**
   * Display the results of the counted votes
   *
   */
  public void displayResults() {
    if (!votes_counted)
      return;
    
    new VoteCounterDisplay(my_election, vote_counts, write_ins);
  }
}
