/*
 * ElectionBuilder.java
 * Last Revised:  26 Oct 97
 * Ben Adida (ben@mit.edu)
 */

package evote.election.builder;

import evote.util.quick_message.*;
import evote.election.*;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

import evote.anonymizer.*;
import evote.counter.*;
import evote.admin.*;
import evote.voter.*;
import evote.commissioner.*;

/**
 *  The ElectionBuilder is a standalone java application that creates
 *  an Election object using the information specified by the user.
 *  A file containing an Election object is either created or opened
 *  and the details of the ballot are entered by the user.
 */
public class ElectionBuilder extends Frame implements ActionListener {
	
    private Election myElection;
    private String directory, file;

    /* Parameter Objects */
    /* stored in Election object */

    /* Setup Panels */
    private VoterSetupPanel voterPanel;
    private CounterSetupPanel counterPanel;
    private AdminSetupPanel adminPanel;
    private AnonymizerSetupPanel anonymizerPanel;
    private CommissionerSetupPanel commissionerPanel;
    private QuestionUserSetupPanel questionUserPanel;
    private KeySetupPanel keyPanel;
    private TimeSetupPanel timePanel;

    /* Menu */
    private Menu electionMenu;
  
    // Election Menu Items
    private MenuItem loadFileItem, saveFileItem, exitItem;
  
    // Setup Menu
    private Menu setupMenu;

    // Setup menu items
    private MenuItem voterItem, counterItem, adminItem, anonymizerItem, commissionerItem, keysItem, timesItem, questionUserItem;
    
    // static final String
    private static final String VOTER_SETUP= "Voter";
    private static final String ADMIN_SETUP= "Administrator";
    private static final String COUNTER_SETUP= "Counter";
    private static final String ANONYMIZER_SETUP= "Anonymizer";
    private static final String COMMISSIONER_SETUP= "Commissioner";
    private static final String KEYS_SETUP= "Cryptographic Keys";
    private static final String TIMES_SETUP= "Voting Times";
    private static final String QUESTION_USER_SETUP= "Questions and Users";
  
    // Main Panel and Card Layout
    Panel mainPanel;
    CardLayout myCardLayout;

    public ElectionBuilder() {
	createGUI();
	myElection= new Election();
	init(myElection);
    }

    public ElectionBuilder(Election elec) {
	createGUI();
	init(elec);
    }

    private void createGUI() {

	/* Menu */
	MenuBar myMenuBar= new MenuBar();
    
	// Election Menu
	electionMenu= new Menu("Election");
    
	loadFileItem= new MenuItem("Load Election");
	saveFileItem= new MenuItem("Save Election");
	exitItem= new MenuItem("Exit");

	electionMenu.add(loadFileItem);
	electionMenu.add(saveFileItem);
	electionMenu.add(exitItem);

	loadFileItem.addActionListener(this);
	saveFileItem.addActionListener(this);
	exitItem.addActionListener(this);
    
	myMenuBar.add(electionMenu);

	// set up the Menu Items
	voterItem= new MenuItem(VOTER_SETUP);
	voterItem.addActionListener(this);

	counterItem= new MenuItem(COUNTER_SETUP);
	counterItem.addActionListener(this);

	adminItem= new MenuItem(ADMIN_SETUP);
	adminItem.addActionListener(this);

	anonymizerItem= new MenuItem(ANONYMIZER_SETUP);
	anonymizerItem.addActionListener(this);

	commissionerItem= new MenuItem(COMMISSIONER_SETUP);
	commissionerItem.addActionListener(this);

	keysItem= new MenuItem(KEYS_SETUP);
	keysItem.addActionListener(this);

	timesItem= new MenuItem(TIMES_SETUP);
	timesItem.addActionListener(this);

	questionUserItem= new MenuItem(QUESTION_USER_SETUP);
	questionUserItem.addActionListener(this);

	setupMenu= new Menu("Setup");
	setupMenu.add(voterItem);
	setupMenu.add(counterItem);
	setupMenu.add(adminItem);
	setupMenu.add(anonymizerItem);
	setupMenu.add(commissionerItem);
	setupMenu.add(keysItem);
	setupMenu.add(timesItem);
	setupMenu.add(questionUserItem);

	myMenuBar.add(setupMenu);

	this.setMenuBar(myMenuBar);
    
	// Main Panel
	mainPanel= new Panel();
	myCardLayout= new CardLayout();
	mainPanel.setLayout(myCardLayout);

	this.add("Center",mainPanel);
    
	this.pack();
	this.setSize(550,650);
	this.show();
    }

    /* set up the panels with the right stuff */
    private void init(Election elec) {
	/* currently, we assume that's what's returned
	 * by the Election object is non-null
	 */

	myElection= elec;

	voterPanel= new VoterSetupPanel(myElection.getVoterParameters());
	counterPanel= new CounterSetupPanel(myElection.getCounterParameters());
	adminPanel= new AdminSetupPanel(myElection.getAdminParameters());
	anonymizerPanel= new AnonymizerSetupPanel(myElection.getAnonymizerParameters());
	commissionerPanel= new CommissionerSetupPanel(myElection.getCommissionerParameters());

	/* clear the card layout */
	mainPanel.removeAll();

	/* add the new panels */
	mainPanel.add(VOTER_SETUP, voterPanel);
	mainPanel.add(COUNTER_SETUP, counterPanel);
	mainPanel.add(ADMIN_SETUP, adminPanel);
	mainPanel.add(ANONYMIZER_SETUP, anonymizerPanel);
	mainPanel.add(COMMISSIONER_SETUP, commissionerPanel);
    
	/* Recreate the other panels */
	timePanel= new TimeSetupPanel(myElection.getCounterParameters(),
				      myElection.getAdminParameters(),
				      myElection.getAnonymizerParameters(),
				      myElection.getCommissionerParameters());

	keyPanel= new KeySetupPanel(myElection.getCounterParameters(),
				    myElection.getAdminParameters(),
				    myElection.getAnonymizerParameters(),
				    myElection.getCommissionerParameters());

	questionUserPanel= new QuestionUserSetupPanel(myElection);

	/* Add these older panels */
	mainPanel.add(TIMES_SETUP, timePanel);
	mainPanel.add(KEYS_SETUP, keyPanel);
	mainPanel.add(QUESTION_USER_SETUP, questionUserPanel);
    }

    public void actionPerformed(ActionEvent ae) {
	Object source= ae.getSource();

	if (source == saveFileItem) {
	    try {
		FileDialog fd= new FileDialog(this);
		if ((this.file != null) && (this.directory!= null)) {
		    fd.setDirectory(this.directory);
		    fd.setFile(this.file);
		}
		fd.show();
		this.directory= fd.getDirectory();
		this.file= fd.getFile();
		FileOutputStream fos= new FileOutputStream(new File(this.directory,this.file));
		ObjectOutputStream oos= new ObjectOutputStream(fos);
		oos.writeObject(myElection);
		oos.close();
	    } catch(Exception exc) {
		System.out.println("coudln't save file : " + exc.toString());
	    }
	    return;
	}

	/* Load menu item */
	if (source == loadFileItem) {
	    try {
		FileDialog fd= new FileDialog(this);
		fd.show();
		this.directory= fd.getDirectory();
		this.file= fd.getFile();
		FileInputStream fis= new FileInputStream(new File(this.directory,this.file));
		ObjectInputStream ois= new ObjectInputStream(fis);
		Election tempElection= (Election)ois.readObject();
		ois.close();
	
		init(tempElection);
	    } catch(Exception exc) {
		System.out.println("couldn't load file :  " + exc.toString());
	    }
	    return;
	}

	/* Quit menu item */
	if (source == exitItem) {
	    this.dispose();
	    System.exit(0);
	}

	// at this point, it's only menu items
	// in the Setup Menu
	if (source instanceof MenuItem) {
	    myCardLayout.show(mainPanel, ((MenuItem)source).getLabel());
	    return;
	}

    }

    public static void main(String argv[]) {
	new ElectionBuilder();
    }
}
