/* E-Voting Project
 * Secure Connection
 */

package evote.network.server;

import java.io.*;
import java.net.*;
import evote.crypto.*;
import evote.network.server.*;
import evote.lang.*;
import java.math.*;
import evote.util.genrand.*;

/**
 * A Secured Connection.
 *
 * This is an extension to <pre>Connection</pre> which encrypts communication
 * on the network. After the SecureConnection is created, encryption and
 * decryption are transparent.
 *
 * @see evote.network.server.CryptMessage
 */
public class SecureConnection extends Connection {
  private RsaKeys server_key;
  private EncryptClass session_encrypter;
  private EncryptClass server_encrypter;
  private IntegrityClass integrity_checker;

  private static final int SESSION_KEY_SIZE= 10;

  /* This is for Connections on top of Connections */
  private boolean is_layer= false;
  private Connection lower_connection= null;

  /* This is whether or not we accept insecure connections */
  private boolean accept_insecure= false;
  private boolean is_insecure= false;

  /* Is this a server */
  private boolean is_server= false;

  /**
   * Constructor for client-side secure connections
   *
   * This constructs a new client-side <pre>SecureConnection</pre>.
   *
   * @argument server The string of the hostname to connect to.
   * @argument port the port of the server to connect to.
   * @argument gr The random number generator
   * @argument rsk The RSA Key Pair that performs the encryption of the session key. No need for the private part of the key pair.
   * @return The created SecureConnection
   * @exception evote.network.server.ConnectionException
   *    If the connection cannot be created for whatever reason, this
   *    exception is thrown
   */
  public SecureConnection(String server, int port, GenRand gr, RsaKeys rsk) throws ConnectionException {
    super(server,port);
    server_key= rsk;
    server_encrypter= new RsaEncryption(server_key);

    /* here is where we make our encrypter */
    byte[] ses_key = gr.nextBytes(SESSION_KEY_SIZE);
    /* Make Session Key Positive */
    ses_key[0] &= 0x7f;
    session_encrypter= new Blowfish(ses_key);

    /* here is where we make our MAC class */
    /* We seed it with the GenRand */
    integrity_checker= new ShaMACIntegrity(gr);

    doSecureClientHandshake();
  }

  /** 
   * The server-side constructor for <pre>SecureConnection</pre>
   * 
   * This constructs a SecureConnection given a socket. Thus, the purpose
   * is mainly for server-side. The <pre>SecureServer</pre> takes care 
   * of creating these, so the user should not bother with this constructor.
   *
   * @argument s the socket
   * @argument rsk the RSA Key Pair. Must contain the private key.
   * @argument gr a Random number generator
   * @argument accept_insec Whether or not this secureconnection, on the 
   * server side, accepts insecure messages.
   * @return the SecureConnection
   * @exception evote.network.server.ConnectionException
   *    this is thrown if anything goes wrong in creating the connection.
   */
  public SecureConnection(Socket s, RsaKeys rsk, GenRand gr, boolean accept_insec) throws ConnectionException {
    super(s);
    is_server= true;
    this.server_key= rsk;

    /* We make the Server Encrypter */
    server_encrypter= new RsaEncryption(rsk);

    /* Determine if we accept insecure connections */
    this.accept_insecure= accept_insec;

    doSecureServerHandshake();
  }

  /**
   * Creates a <pre>SecureConnection</pre> layered over another
   * SecureConnection, for double/triple/n encryption purposes.
   *
   * @argument c the SecureConnection to layer on top of.
   * @argument gr The Random Number Generator
   * @argument rsk The RSA Key Pair used to encrypt messages at this
   * layer
   * @return the SecureConnection
   * @exception evote.network.server.ConnectionException
   *    this is thrown if anything goes wrong in creating the connection.
   */
  public SecureConnection(SecureConnection c, GenRand gr, RsaKeys rsk) throws ConnectionException {
    super();
    this.is_layer= true;
    this.lower_connection= c;
    this.server_key= rsk;
    
    this.accept_insecure= c.accept_insecure;
    this.is_insecure= c.is_insecure;

    this.is_server= c.is_server;
    
    server_encrypter= new RsaEncryption(rsk);
    
    /* If we're not the server, then we need to set the
     * Session encrypter, and integrity checker
     */
    if (!this.is_server) {
      /* here is where we make our encrypter */
      byte[] ses_key = gr.nextBytes(SESSION_KEY_SIZE);
      /* Make Session Key Positive */
      ses_key[0] &= 0x7f;
      session_encrypter= new Blowfish(ses_key);
      
      /* here is where we make our MAC class */
      /* We seed it with the GenRand */
      integrity_checker= new ShaMACIntegrity(gr);
    }
  }
    

  /* The Next two methods perform the handshaking and session key
   * exchange. The protocol works this way:
   * Currently, no handshaking happens.
   */

  /* Client handshake */
  private void doSecureClientHandshake() {

    /* We no longer do any handshaking, we send the key in every message */

  }

  /* Server Handshake */
  private void doSecureServerHandshake() {
    
    /* We no longer do any handshaking. See above */
    
  }

  /** 
   * This sends a message, encrypted.
   *
   * @argument m The message to be sent
   * @exception evote.network.server.ConnectionException
   *    If sending a message goes wrong, this exception is thrown.
   */
  public void sendMessage(Message m) throws ConnectionException {
    
    /* If this connection is insecure, then send the Message straight */
    if (is_insecure) {
      super.sendMessage(m);
      return;
    }
    
    if (session_encrypter==null)
      throw(new ConnectionException("No Session Key Yet"));

    try {
      //System.out.println("Creating a CryptMessage");
      CryptMessage cm= new CryptMessage(m, session_encrypter, server_encrypter, integrity_checker);
      
      /* Here we check to see if this is a Layer 
       * If we are layering connections, we pass this down to the
       * lower level, otherwise, we just send the CryptMessage
       */
      if (this.is_layer)
	lower_connection.sendMessage(cm);
      else
	super.sendMessage(cm);
    }
    catch(CryptoException ce) {
      throw(new ConnectionException("Encryption problem: " + ce.toString()));
    }

  }

  /**
   * Sends a message unencrypted
   *
   * This is used to send a message unencrypted.
   * This is especially useful for layered secure connections
   * in which the other end expects a doubly encrypted message
   * where only a singly-encrypted acknowledgement can be
   * sent by an intermediary
   *
   * @argument m the message to send.
   * @exception evote.network.server.ConnectionException
   *    if anything goes wrong, this exception is thrown.
   */
  public void sendUnencryptedMessage(Message m) throws ConnectionException {
    
    if (this.is_layer) {
      lower_connection.sendUnencryptedMessage(m);
      return;
    }

    super.sendUnencryptedMessage(m);
  }

  /** 
   * This blocks until a message is received from the Connection.
   * If the message received is not a CryptMessage, it is returned
   * unmodified, and a flag is set to declare this connection insecure.
   * If the message received is a CryptMessage, it is decrypted.
   *
   * @return the decrypted Message
   * @exception evote.network.server.ConnectionException
   *    This exception is thrown if anything goes wrong in the process
   *    of receiving the message.
   */
  public Message receiveMessage() throws ConnectionException {
    
    Message m;

    /* If we are layering Connections, we call receiveMessage
     * on the Lower level connection
     */
    if (this.is_layer) 
      m= lower_connection.receiveMessage();
    else
      m= super.receiveMessage();

    /* We check what kind of message m is */
    if (!(m instanceof CryptMessage)) {
      /* If we are the Client, and we are a layered SecureConnection,
       * we can accept the non-Crypt message */
      if ((!this.is_server)&&(this.is_layer))
	return(m);
      
      if (accept_insecure) {
	is_insecure= true;
	return(m);
      } else {
	throw(new ConnectionException("Insecure Connection Attempt Not Accepted"));
      }
    }

    /* If this is the server, and no key has yet been
     * received, we set the session_encrypter 
     */
    if (is_server) {
      try {
	byte[] ses_key= ((CryptMessage)m).getEncryptedSessionKey();
	//System.out.println(" Got session Key ");
	if (ses_key==null)
	  System.out.println("ses_key NULL!");
	if (server_encrypter==null)
	  System.out.println("server_encrypter NULL!");
	ses_key= server_encrypter.decrypt(ses_key);
	//System.out.println(" Decrypted Session Key ");
	session_encrypter= new Blowfish(ses_key);
	//System.out.println(" Made Session Encrypter ");

	/* Make the integrity Checker */
	ByteArrayInputStream bais= new ByteArrayInputStream(session_encrypter.decrypt(((CryptMessage)m).getEncryptedCommitKeys()));
	ObjectInputStream ois= new ObjectInputStream(bais);
	integrity_checker= new ShaMACIntegrity((ShaCommitmentKeys)ois.readObject());

	// System.out.println(" Got Commitment Keys ");

	ois.close();
	bais.close();
      }
      catch(Exception e) {
	throw(new ConnectionException("couldn't get session key: " + e.toString()));
      }
    }

    try {
      Message decrypt_m= ((CryptMessage)m).decryptAndVerify(session_encrypter, integrity_checker);
      return(decrypt_m);
    }
    catch(CryptoException ce) {
      ce.printStackTrace();
      throw(new ConnectionException("Encryption Problem : " + ce.toString()));
    }

  }

}
