/*
 *  Anonymizer.java
 *  Last Revised:  17 Jan 1999
 *  Randy L. Milbert <rmilbert@mit.edu>
 *
 *  completely reworked by
 *  Ben Adida (ben@mit.edu)
 */

package evote.anonymizer;

import evote.network.server.*;
import evote.util.genrand.*;
import evote.util.shuffle.*;
import evote.util.constants.*;
import evote.crypto.*;
import evote.commissioner.complaints.*;
import evote.network.message.*;
import evote.util.quick_message.*;
import evote.election.*;
import evote.store.*;

import java.applet.*;
import java.awt.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.lang.*;
import java.math.BigInteger;


/**
  The AnonymizerServer contains code for handling connections
  and saving the objects received to disk.  The order of those
  objects will eventually be shuffled and they will be sent
  on to their final destination.
  */
public class AnonymizerServer implements ConnectionHandler {
    Server my_server;
    ObjectStore myStore;
    private int num_votes = 0;
    private GenRand RNG;
    RsaEncryption anonSigner;
    RsaKeys commisKeys;
    String commisServer;
    public  String my_name = "Anonymizer";
    public  int error_type = 0;

    private boolean relay_mode= true;

    /* Parameters */
    Election myElection;

    public AnonymizerServer(Election elec) {
	RNG= new GenRand("evote.util.genrand.MouseGenerator");
	RNG.seed();

	/* Load up the parameters */
	myElection= elec;
	if (myElection == null) {
	    /* BIG ERROR */
	    System.out.println("problem: null election");
	    System.exit(0);
	}

	Constants.init(myElection);

	anonSigner= new RsaEncryption(Constants.anonKeys);

	try {
	    myStore= ObjectStore.newInstance("evote.store.DiskObjectStore");
	} catch (Exception exc) {
	    System.err.println(exc.toString());
	    System.exit(0);
	}

	ObjectStoreParameters osp= new ObjectStoreParameters();
	osp.setParameter("directory", "./");

	try {
	    myStore.init(osp);
	} catch (Exception exc) {
	    System.err.println(exc.toString());
	    System.exit(0);
	}

	try {
	    my_server= new SecureServer(my_name, Constants.anonReceivePort, this, Constants.anonNumThreads, Constants.anonQueueSize, Constants.anonKeys, RNG, false);
	    my_server.setTimes(Constants.anonStartDate, Constants.anonEndDate);
	    my_server.start();
	}
	catch(Exception e) {
	    new QuickMessage("Error:  Could Not Start Server",
			     "Error:  an error occurred while trying to start the server.");
	    System.out.println("exception: " + e.toString());
	    System.exit(0);
	}

    }

    public void startDemo() {
	my_server.startDemo();
    }

    public void stopDemo() {
	my_server.stopDemo();
    }
	
    public static void main(String argv[]) {
    
	try {
	    ElectionLoader eLoader= new ElectionLoader(argv);
	    Election myElec= eLoader.getElection();
      
	    new AnonymizerServer(myElec);
	    System.out.println("Anon Server starting ....");
	} catch (Exception exc) {
	    System.out.println("Error happened: " + exc.toString());
	}
    }

    public void handleConnection(Connection c) {

	try {
      
	    my_server.setStatus("Received\nConnection");

	    Message msg = c.receiveMessage();
      
	    if (relay_mode) {
		try {
		    my_server.setStatus("Relaying\nVote");
		    Connection con= new Connection(Constants.counterServer, Constants.counterReceivePort);
		    con.sendMessage(msg);
		    con.sendMessage(new DoneMessage());

		    con.close();

		    my_server.incrementCount();
		    my_server.clearStatus();
		} catch(Exception exc) {
		    System.err.println("error: " + exc.toString());
		}
	
	    } else {
		try {
		    my_server.setStatus("Storing\nVote");
		    myStore.storeObject(myStore.getNextKey(), msg);
		    my_server.incrementCount();
		    my_server.clearStatus();
		} catch(ObjectStoreException exc) {
		    System.err.println(exc.toString());
		}
	    }
      
	    my_server.setStatus("Confirming\nwith Voter");
	    OKMessage okm = new OKMessage(my_name);
	    c.sendUnencryptedMessage(okm);
	    my_server.clearStatus();
	} catch (Exception exp) {
	    Complaint comp = new Complaint(my_name, (c.getInetAddress()).toString(), 0, exp.toString(), RNG);
	    ComplaintMessage compMsg = new ComplaintMessage(comp, anonSigner);
	    my_server.log(new LogEntry(LogEntry.WARNING_LOG, my_name + "\n" + (c.getInetAddress()).toString() + "\n" + error_type + "\n" + exp.toString()));
	    try {
		SecureConnection commisConnection = new SecureConnection(commisServer, Constants.commisReceivePort, RNG, commisKeys);
		commisConnection.sendMessage(compMsg);
		c.sendUnencryptedMessage(compMsg);
	    } catch (Exception e) {
		my_server.log(new LogEntry(LogEntry.SERIOUS_LOG, "Can't connect to commissioner."));
	    }
	}
    }

    public void handleError(ConnectionError ce) {
	try {
	    Complaint comp = new Complaint(my_name, "", 0, ce.toString(), RNG);
	    ComplaintMessage compMsg = new ComplaintMessage(comp, anonSigner);
	    my_server.log(new LogEntry(LogEntry.WARNING_LOG, my_name + "\n" + "" + "\n" + error_type + "\n" + ce.toString()));
	    SecureConnection commisConnection = new SecureConnection(commisServer, Constants.commisReceivePort, RNG, commisKeys);
	    commisConnection.sendMessage(compMsg);
	} catch (Exception exp) {
	    my_server.log(new LogEntry(LogEntry.SERIOUS_LOG, "Can't connect to commissioner."));
	}
    }

    /* server shutting down */
    public void serverShutdown() {
	if (relay_mode)
	    return;
    
	Vector badVotes= new Vector();

	// now we connect to the counter and send everything at once
	// for now this is in the same order as what we had, but we
	// can randomize it later
    
	// connect
	try {
	    Enumeration all_msgs= myStore.loadAllObjects();
      
	    my_server.setStatus("Connecting\nto Counter");
	    System.out.println("ANONYMIZER: CONNECTING to Counter");
	    Connection con= new Connection(Constants.counterServer, Constants.counterReceivePort);

	    int counter= 0;
	    Message nextMessage= null;

	    my_server.setStatus("Sending\nvotes");
	    while (all_msgs.hasMoreElements()) {
		try {
		    if ((counter % 10) == 0) {
			System.out.println("resetting connection");
			con.sendMessage(new DoneMessage());
			con= new Connection(Constants.counterServer, Constants.counterReceivePort);
		    }
	  
		    System.out.println("sending out ONE VOTE TO THE COUNTER");
		    nextMessage= (Message)all_msgs.nextElement();
		    con.sendMessage(nextMessage);
		    counter+= 1;
		} catch(Exception exc) {
		    System.err.println(exc.toString());
		    badVotes.addElement(nextMessage);
		}
	    }

	    my_server.setStatus("done");
	    System.out.println("ANONYMIZER: SENDING Done Message to Counter");
	    con.sendMessage(new DoneMessage());
	} catch(Exception exc) {
	    System.err.println(exc.toString());
	    exc.printStackTrace();
	}

	try {
	    System.out.println("There are " + badVotes.size() + " bad votes");
	    FileOutputStream fos= new FileOutputStream("BAD_VOTES.txt");
	    ObjectOutputStream oos= new ObjectOutputStream(fos);
      
	    oos.writeObject(badVotes);

	    oos.close();
	    fos.close();
	} catch(Exception exc) {
	    System.err.println("Problem writing out bad votes!");
	    exc.printStackTrace();
	}
    }
}

