/* Electronic Voting Project
 * Counter
 * written by Ben Adida (ben@mit.edu)
 */

package evote.counter;

import java.math.*;
import java.io.*;
import java.net.*;
import java.util.Vector;
import java.util.Enumeration;
import evote.crypto.*;
import evote.util.genrand.*;
import evote.util.constants.*;
import evote.lang.*;
import evote.network.message.*;
import evote.network.server.*;
import evote.counter.*;
import evote.util.constants.*;
import evote.election.*;
import evote.commissioner.complaints.*;

/**
 * The Counter Server
 *
 * This server is the Counter in the Voting system. It receives votes,
 * verifies that they are properly signed, properly committed, and not
 * already received. It also checks that the votes correspond
 * to the right election.
 * <p>
 * If all these tests pass, the vote is kept in memory. 
 * <p>
 * Once all votes have been received, they can be counted.
 * They can also be looked at individually. They are, of course, unlinked to
 * an identity.
 * <p>
 * The Counter uses the SecureServer layer.
 */

public class Counter implements ConnectionHandler {
    private Server my_server;
    private VoteList my_votelist;
    private Signer admin_signer, counter_signer;
    private RsaKeys adminKeys;
    private Commitment vote_commitment;
    private Vector my_vote_listeners;
    private RsaKeys my_keys;
    private GenRand my_genrand;

    private int attempted_votes= 0;
  
    /**
     * This creates a new Counter Server
     * 
     * @argument elec The Election object that describes the Election.
     */
    public Counter(Election elec) {
    
	/* Init the constants */
	Constants.init(elec);
	CounterParameters counterParams= elec.getCounterParameters();

	adminKeys= new RsaKeys(Constants.adminPublic, Constants.adminMod);

	my_keys= new RsaKeys(Constants.counterPublic, Constants.counterPrivate, Constants.counterMod);
	counter_signer= new RsaEncryption(my_keys);

	my_vote_listeners= new Vector();
    
	/* Add the Interface as a Vote Listener */
	this.addVoteListener(new CounterInterface(elec,this));

	my_genrand= new GenRand();
	my_genrand.seed();

	try {
	    my_server= new SecureServer("Counter",Constants.counterReceivePort, this, Constants.counterNumThreads, Constants.counterQueueSize, my_keys, my_genrand);
	    ((SecureServer)my_server).setAcceptInsecure(true);
	    my_server.setTimes(counterParams.getStart(), counterParams.getEnd());
	    my_server.start();
	}
	catch(ServerException e) {
	    System.err.println("Server could not start : " + e.toString());
	    System.exit(0);
	}
	admin_signer= new RsaEncryption(adminKeys);
	my_votelist= new VoteList();
    }
  
    public void startDemo() {
	my_server.startDemo();
    }

    public void stopDemo() {
	my_server.stopDemo();
    }

    private void sendAck(Message m, Connection c) {
	/* Here, should get the id out of the message m,
	 * and return the proper acknowledgement
	 */
	try {
	    c.sendUnencryptedMessage(new OKMessage());
	}
	catch(ConnectionException e) {
	    /* log this error */
	    my_server.log(new LogEntry(LogEntry.SERIOUS_LOG,"Could not send acknowledgement: " + e.toString()));
	}
    }
  
    private void sendError(Message m, Connection c) {
	/* Here we send an error message back
	 */
	//c.sendUnencryptedMessage(new ErrorMessage());
    }

    /** 
     * Run the Counter Server standalone
     */
    public static void main(String argv[]) {
	try {
	    ElectionLoader eLoader= new ElectionLoader(argv);
	    Election myElec= eLoader.getElection();
      
	    new Counter(myElec);
	}
	catch(Exception exc) {
	    System.out.println("couldn't run Counter: " + exc.toString());
	}
    }

    /***************************/
    /* Handling of Vote Events */
    /***************************/

    public void addVoteListener(VoteListener vl) {
	my_vote_listeners.addElement(vl);
    }

    public void removeVoteListener(VoteListener vl) {
	my_vote_listeners.removeElement(vl);
    }

    private void notifyVoteListeners(Vote v) {
	VoteEvent ve= new VoteEvent(v);
    
	Enumeration enum= my_vote_listeners.elements();
	while (enum.hasMoreElements()) {
	    // System.out.println("Notifying a Vote Listener");
	    ((VoteListener)enum.nextElement()).receiveVote(ve);
	}
    }

    /***************/

    public void handleConnection(Connection c) {
	Message incoming;
	boolean flag;

	try {
	
	    my_server.setStatus("Received\nConnection");

	    my_server.setStatus("receiving\nvotes");

	    while(true) {
		System.out.println("COUNTER: waiting for Message");
		incoming= c.receiveMessage();
		System.out.println("COUNTER: received a Message");
	
		if (incoming instanceof VoteMessage) {
		    System.out.println("COUNTER: handling Vote");
		    flag= handleVote(((VoteMessage)incoming).getVote());
		    System.out.println("COUNTER: done handling Vote");
		    my_server.incrementCount();
		}
	
		if (incoming instanceof DoneMessage) {
		    my_server.setStatus("done\nreceiving\nvotes");
		    System.out.println("COUNTER: done receiving votes");
		    break;
		}
	
		System.out.println("COUNTER: end of one receipt iteration");
	    }

	    System.out.println("COUNTER: out of receipt loop");
	} catch(Exception exc) {
	    System.out.println("COUNTER exception");
	    System.out.println(exc.toString());
	    exc.printStackTrace();
	} finally {
	    my_server.clearStatus();
	}
    
	System.out.println("COUNTER: done handling CONNECTION");

	try {
	    c.close();
	} catch(Exception exc) {
	}
    }
  
    public void handleError(ConnectionError e) {
	my_server.log(new LogEntry(LogEntry.WARNING_LOG,e.toString()));
    }

    public synchronized void reallyQuit() {
	notifyAll();
    }

    public synchronized void serverShutdown() {
	try {
	    wait();
	}
	catch(InterruptedException exc) {
	}
    }
  
    private synchronized int incrementVoteCount() {
	attempted_votes+= 1;
	return(attempted_votes);
    }

    private synchronized boolean handleVote(Vote v) {
    
	int vote_num= this.incrementVoteCount();
    
	try {
	    /* Check the validity of the signature */

	    /* Log access */
	    my_server.log(new LogEntry(LogEntry.NORMAL_LOG,"Received a Vote #" + vote_num));

	    /* Log stuff about the Vote */
	    try {
		my_server.log(new LogEntry(LogEntry.NORMAL_LOG,"Vote #" + vote_num + " commited bytes\n " + new BigInteger(v.getCommittedBytes()) + "\n\n Signature \n " + new BigInteger(v.getAdminSignature()) ));
	    }
	    catch(Exception e) {
		/* The bytes aren't present */
		my_server.log(new LogEntry(LogEntry.SERIOUS_LOG,"Vote #" + vote_num + "\n doesn't contain the committed bytes or admin signature"));
	    }

	    if (!v.verifyAdminSign(admin_signer)) {
		my_server.log(new LogEntry(LogEntry.SERIOUS_LOG,"Admin Signature is Invalid for Vote #" + vote_num + " !"));
	
		/* Send a complaint */
		new ComplaintSender(new Complaint("Counter","Voter", Complaint.Connection_Error, "could not receive a Message from the anonymizer", my_genrand), my_genrand, counter_signer, my_server);


		return(false);
	    }

	    /* Check the commitment of the Vote */
	    if (!v.verifyCommitment(new ShaCommitment((ShaCommitmentKeys)v.getCommitmentKeys()))) {
		my_server.log(new LogEntry(LogEntry.SERIOUS_LOG,"Commitment is Invalid for Vote #" + vote_num + " !"));	

		/* Send a complaint */
		new ComplaintSender(new Complaint("Counter","Voter", Complaint.Connection_Error, "could not receive a Message from the anonymizer", my_genrand), my_genrand, counter_signer, my_server);

		return(false);
	    }
      
	    /* Check validity of Question List */
	    ShaCommitmentKeys ql_keys= new ShaCommitmentKeys(new BigInteger("0"),new BigInteger("0"));
	    if (!v.verifyQuestionList(new ShaCommitment(ql_keys), admin_signer)) {
		/* Log Error */
		my_server.log(new LogEntry(LogEntry.SERIOUS_LOG,"Question List signature is invalid for Vote #" + vote_num));

		/* Send a complaint */
		new ComplaintSender(new Complaint("Counter","Voter", Complaint.Questionlist_Invalid, "The signature on the QuestionList is not valid", my_genrand), my_genrand, counter_signer, my_server);
	
		return(false);
	    }

	    /* Log Success of Vote */
	    my_server.log(new LogEntry(LogEntry.NORMAL_LOG,"Vote #" +vote_num + " passed"));
      
	    /* Add to vote list */
	    my_votelist.addVote(v);

	    /* If Vote added successfully, notify listeners */
	    notifyVoteListeners(v);

	}
	catch(Exception e) {
	    System.err.println("COUNTER: Error recording vote!");
	    e.printStackTrace();
	    my_server.log(new LogEntry(LogEntry.SERIOUS_LOG,"Vote #" + vote_num + " had already been received!"));

	    /* Send a complaint */
	    new ComplaintSender(new Complaint("Counter","Voter", Complaint.Connection_Error, "could not receive a Message from the anonymizer", my_genrand), my_genrand, counter_signer, my_server);


	    return(false);
	}
	return(true);
    }
  
}


class VoteList {
    private Vector[] table;
    private boolean[] initialized;
    private ObjectOutputStream myOutputStream;

    private VoteList(int max) {
	File test_f= null;
	int counter= 0;
	while(true) {
	    test_f= new File("VOTES-" + counter + ".txt");
	    if (!test_f.exists())
		break;
	    counter+= 1;
	}

	try {
	    FileOutputStream fos= new FileOutputStream(test_f);
	    myOutputStream= new ObjectOutputStream(fos);
	} catch (IOException exc) {
	    System.err.println("error opening outputstream");
	    System.exit(0);
	}

	table= new Vector[max];
	initialized= new boolean[max];
	for (int i=0; i<max; i++) 
	    initialized[i]= false;
    }
  
    public VoteList() {
	this((int)Math.pow(2,16));
    }

    private int getVoteHash(Vote v) {
	byte[] b;
	try {
	    b= v.getCommittedBytes();
	}
	catch(Exception e) {
	    return(0);
	}
	byte[] short_b= new byte[2];
	for (int i=0; i<2; i++) {
	    // System.out.println("byte #" + i + " ; " + b[i]);
	    short_b[i]= b[i+1];
	}

	int index= (int) ((new Byte(b[0])).intValue() * Math.pow(2,8) + (new Byte(b[1])).intValue());
	index= Math.abs(index);
	return(index);
    }

    private boolean byteArrayCompare(byte[] a, byte[] b) {
	if (a.length!=b.length) return(false);
	for (int i= a.length-1; i>=0; i--) {
	    if (a[i]!=b[i])
		return(false);
	}
	return(true);
    }

    private boolean search(Vector vect, Vote v) {
	if (vect==null)
	    return(false);

	Enumeration enum= vect.elements();
	try {
	    while (enum.hasMoreElements()) {
		if (byteArrayCompare(v.getCommittedBytes(),((Vote)enum.nextElement()).getCommittedBytes()))
		    return(true);
	    }
	}
	catch(Exception e) {
	    /* Should NOT happen */
	}
	return(false);
    }

    public synchronized void addVote(Vote v) throws Exception {
	int vote_hash= getVoteHash(v);
    
	// System.out.println("Got Vote Hash: " + vote_hash);

	if (search(table[vote_hash],v)) {
	    // System.out.println("vote HERE already!");
	    throw(new Exception("Vote Already here"));
	}

	// System.out.println("Vote hasn't been here yet");

	if (!initialized[vote_hash]) {
	    table[vote_hash]= new Vector();
	    initialized[vote_hash]= true;
	    // System.out.println("initialized part of table");
	}

	// System.out.println("adding element");
	table[vote_hash].addElement(v);
	// System.out.println("element added");

	// Add vote to output file
	myOutputStream.writeObject(v);
	myOutputStream.flush();

    }

    public boolean find(Vote v) {
	return(search(table[getVoteHash(v)],v));
    }

}
