/* Electronic Voting Project
 * Mark A. Herschberg (hershey@mit.edu)
 */

package evote.crypto;

import java.math.BigInteger;
import java.security.*;
import java.util.Random;
import java.io.*;

import evote.util.constants.*;
import evote.util.genrand.*;

/**
  This class defines an RSA key.  It consists of BigIntegersp, q, e, d, and n.
  Keys with some fields missing can also be created.
  */
public class RsaKeys implements Serializable {
  public BigInteger p = null;
  public BigInteger q = null;
  public BigInteger e = null;
  public BigInteger d = null;
  public BigInteger n = null;
  
  /**
    Creates an empty RSA Key.
    */
  public RsaKeys() {}
  
  /**
    Creates a RSA key by using the byte array as the source for SecureRandom,
    which is used to generate BigInteger primes.  The seed itself is rehashed
    using MessageDigest's SHA to generate new seeds until appropriate primes
    are created.
    */
  public RsaKeys(byte[] seed) {
    BigInteger eCheck;
    byte[] tempseed = seed;
    /* tempseed keeps gettign rehashed to generate new seeds for p and q, until
       an appropriate p and q are created */
    do {
      try {
	tempseed = (MessageDigest.getInstance("SHA")).digest(tempseed);
      }
      catch (NoSuchAlgorithmException exp) {}
      p = new BigInteger(Constants.SecurityParameter, Constants.certainty, new SecureRandom(tempseed));
      try {
	tempseed = (MessageDigest.getInstance("SHA")).digest(tempseed);
      }
      catch (NoSuchAlgorithmException exp) {}
      q = new BigInteger(Constants.SecurityParameter, Constants.certainty, new SecureRandom(tempseed));
    } while (p.equals(q));
    n = p.multiply(q);
    eCheck = (p.subtract(new BigInteger("1"))).multiply(q.subtract(new BigInteger("1")));
    boolean doagain = true;
    /* now tempseed is rehashed and used to create e and d */
    do {
      do {
	try {
          tempseed = (MessageDigest.getInstance("SHA")).digest(tempseed);
        }
        catch (NoSuchAlgorithmException exp) {}
	e = new BigInteger(Constants.numbits, tempseed);
      } while (!(e.gcd(eCheck)).equals(new BigInteger("1")));
      try {  
	d = e.modInverse(eCheck);
	doagain = false;
      }
      catch (ArithmeticException exp)
	{doagain = true;}
    } while (doagain);
  }
  
  /**
    Creates a RSA key by using a GenRand source as the seed for BigInteger
    primes.
    */
  public RsaKeys(GenRand RNG) {
    BigInteger eCheck;
    Constants cons = new Constants();
    /* GenRand is used to create an appropriate p and q */
    do {
      p = new BigInteger(Constants.SecurityParameter, Constants.certainty, new SecureRandom(RNG.nextBytes(Constants.nextBytes)));
      q = new BigInteger(Constants.SecurityParameter, Constants.certainty, new SecureRandom(RNG.nextBytes(Constants.nextBytes)));
    } while (p.equals(q));
    n = p.multiply(q);
    eCheck = (p.subtract(new BigInteger("1"))).multiply(q.subtract(new BigInteger("1")));
    boolean doagain = true;
    /* GenRand is tapped to create e and d */
    do {
      do {
	e = new BigInteger(Constants.numbits, new SecureRandom(RNG.nextBytes(Constants.nextBytes)));
      } while (!(e.gcd(eCheck)).equals(new BigInteger("1")));
      try {  
	d = e.modInverse(eCheck);
	doagain = false;
      }
      catch (ArithmeticException exp)
	{doagain = true;}
    } while (doagain);
  }
  
  /**
    Creates a RSA key with only e and n specified.
    */
  public RsaKeys(BigInteger pub, BigInteger modulus) {
    p = null;
    q = null;
    e = pub;
    d = null;
    n = modulus;
  }
  
  /**
    Creates a RSA key with only e, d, and n specified.
    */
  public RsaKeys(BigInteger pub, BigInteger priv, BigInteger modulus) {
    p = null;
    q = null;
    e = pub;
    d = priv;
    n = modulus;
  }
  
  /**
    Sets p to the argument.
    */
  public void SetRsaP (BigInteger x) {
    p = x;
  }
  
  /**
    Sets q to the argument.
    */
  public void SetRsaQ (BigInteger x) {
    q = x;
  }

  /**
    Sets e to the argument.
    */
  public void SetRsaE (BigInteger x) {
    e = x;
  }

  /**
    Sets d to the argument.
    */
  public void SetRsaD (BigInteger x) {
    d = x;
  }

  /**
    Sets n to the argument.
    */
  public void SetRsaN (BigInteger x) {
    n = x;
  }

  /**
    Clear out the private stuff
    */
  public void clearPrivate() {
    d= null;
    p= null;
    q= null;
  }

  /**
    Returns p.
    */
  public BigInteger getRsaP () {
    return p;
  }

  /**
    Returns q.
    */
  public BigInteger getRsaQ () {
    return q;
  }

  /**
    Returns e.
    */
  public BigInteger getRsaE () {
    return e;
  }

  /**
    Returns d.
    */
  public BigInteger getRsaD () {
    return d;
  }

  /**
    Returns n.
    */
  public BigInteger getRsaN () {
    return n;
  }

  /**
   * Clone
   */
  public Object clone() {
    RsaKeys otherSelf= new RsaKeys();

    otherSelf.p = this.p;
    otherSelf.q= this.q;
    otherSelf.n= this.n;
    otherSelf.e= this.e;
    otherSelf.d= this.d;

    return(otherSelf);
  }
}





