/* Electronic Voting Project
 * Ben Adida (ben@mit.edu)
 */

/* This class contains a Choice object, and provides ways of
 * encrypting it, signing it, committing it, etc...
 */

package evote.lang;

import java.io.Serializable;
import java.math.*;

import evote.crypto.*;
import evote.question.*;

public class Vote implements Serializable, Cloneable {
  private QuestionList myQuestionList;
  private byte[] questionListSignature;

  private Choice my_choice= null;
  private byte[] choice_bytes= null;
  private byte[] committed_bytes= null;
  private byte[] blinded_bytes= null;
  private byte[] admin_signature= null;
  private byte[] voter_signature= null;

  private CommitmentKeys commit_keys= null;

  private boolean committed= false, voter_signed=false, admin_signed= false,
    blinded= false;
  

  public Vote(Choice c) throws Exception{
    my_choice= c;
    choice_bytes= my_choice.serialize();
  }

  private void checkCommitted() throws Exception {
    if (!committed)
      throw(new Exception("Vote Not Committed"));
  }

  private void checkBlinded() throws Exception {
    if (!blinded)
      throw(new Exception("Vote Not Blinded"));
  }
    
  /* This is for sorting votes only */
  public byte[] getCommittedBytes() throws Exception{
    if (!committed)
      throw(new Exception("Not Committed"));

    return(committed_bytes);
  }

  public byte[] getBlindedBytes() throws Exception {
    if (!blinded)
      throw(new Exception("Not Blinded"));
    
    return(blinded_bytes);
  }

  public byte[] getAdminSignature() throws Exception {
    if (!admin_signed)
      throw(new Exception("Not Signed by Admin"));
    
    return(admin_signature);
  }

  public synchronized void commit(Commitment c) {
    committed_bytes= c.commit(choice_bytes);
    commit_keys= c.getKeys();

    committed= true;

    /* clear out the fields we can't have any longer */
    my_choice= null;

    /* We keep the choice_bytes, because makeAdminSafe takes
     * care of that
     */
  }

  public synchronized void uncommit(Commitment c) throws Exception {

    /* This can only be used for reversible Commitments
     */
	
    /* If the vote is not yet committed, how can we uncommit it!
     */
    checkCommitted();

    try {
      choice_bytes= c.uncommit(committed_bytes);
    }
    catch(Exception e) {
      throw (new Exception("cannot uncommit"));
    }
    committed=false;
  }
    
  public synchronized void adminSign(Signer s) throws Exception {

    /* We check if the vote is committed
     * if it isn't, this procedure call will throw an exception
     */
    checkCommitted();
    checkBlinded();

    /* Check that Administrator hasn't already signed this */
    if (admin_signed)
      throw(new Exception("Already signed by Administrator"));
	
    admin_signature= s.sign(blinded_bytes);
    admin_signed= true;
  }

  public synchronized void blind(Blind b) throws Exception {
    
    /* This deletes the committed_bytes */

    checkCommitted();

    blinded_bytes= b.blind(committed_bytes);

    /* We don't take out the committed bytes, because makeAdminSafe
     * does that 
     */
    /* committed_bytes= null; */
    blinded=true;
  }

  public synchronized void unblind(Blind b) throws Exception {

    /* This restores the committed bytes */
    
    checkBlinded();

    if (admin_signature==null)
      throw(new Exception("not signed"));

    admin_signature= b.unblind(admin_signature);
    blinded=false;
  }

  public synchronized void voterSign(Signer s) throws Exception {

    /* We check if the vote is committed
     */
    checkCommitted();

    /* Check that Voter hasn't already signed this */
    if (voter_signed)
      throw(new Exception("Already signed by Voter"));
		  
    voter_signature= s.sign(committed_bytes);
    voter_signed= true;
  }

  public boolean verifyAdminSign(Signer s) throws Exception {

    checkCommitted();
    if (!admin_signed)
      throw(new Exception("Not signed by administrator"));

    return(s.verify(committed_bytes,admin_signature));
  }

  public boolean verifyVoterSign(Signer s) throws Exception {

    checkCommitted();
    if (!voter_signed)
      throw(new Exception("Not signed by voter"));

    return(s.verify(committed_bytes,voter_signature));
  }

  public boolean verifyCommitment(Commitment c) throws Exception {
    checkCommitted();
    
    if (choice_bytes == null) 
      throw(new Exception("No Choice Bytes"));

    byte[] check_b;
    check_b= c.commit(choice_bytes);

    /* Compare the two arrays */
    if (check_b.length != committed_bytes.length)
      return(false);

    for (int i=0; i<check_b.length; i++)
      if (check_b[i] != committed_bytes[i])
	return(false);
    
    return(true);
  }

  public boolean verifyQuestionList(Commitment c, Signer s) throws Exception {
    
    if ((myQuestionList == null) || (questionListSignature == null))
      throw(new Exception("Question List or signature not present"));
    
    byte[] QLBytes= myQuestionList.serialize();
    ShaCommitmentKeys sha_keys= new ShaCommitmentKeys(new BigInteger("0"), new BigInteger("0"));
    ShaCommitment sha= new ShaCommitment(sha_keys);
    byte[] receivedBytes= sha.commit(QLBytes);
    
    return (s.verify(receivedBytes, questionListSignature));
  }
    
  public Vote makeAdminSafe() throws Exception {
    checkCommitted();
    checkBlinded();

    Vote v2= (Vote)this.clone();
    v2.my_choice= null;
    v2.choice_bytes= null;
    v2.committed_bytes= null;
    v2.commit_keys= null;

    return(v2);
  }

  public Vote makeCounterSafe() throws Exception {
    checkCommitted();
    
    Vote v2= (Vote)this.clone();
    v2.voter_signature= null;
    
    return(v2);
  }

  public void setQuestionList(QuestionList ql, byte[] sign) {
    myQuestionList= ql;
    questionListSignature= sign;
  }

  public QuestionList getQuestionList() {
    return(myQuestionList);
  }

  public byte[] getQuestionListSignature() {
    return(questionListSignature);
  }

  public CommitmentKeys getCommitmentKeys() {
    return(commit_keys);
  }

  public void getSignature(Vote other_v) {
    this.admin_signature= other_v.admin_signature;
    admin_signed=true; //Mark
  }
  
  public Choice getChoice() throws Exception {
    if (my_choice==null)
      my_choice= Choice.deserialize(choice_bytes);
    return(my_choice);
  }

}




